@@ -1,3 +1,4 @@
+import functools
 import torch 
 import torch.nn as nn 
 import torch.nn.functional as F 
@@ -5,86 +6,17 @@ import torch.nn.init as init
 from torch.nn.parameter import Parameter
 import math 
 import numpy as np 
+import torchvision.ops as ops
+from .Base import Model 
+from .Quant import QQuantizers, QObservers, QTYPES
 
-record_params = []
 
 def _resnet_normal(tensor):
 	fan_in, fan_out = init._calculate_fan_in_and_fan_out(tensor)
 	std = math.sqrt(2.0 / float(fan_out))
 	return init._no_grad_normal_(tensor, 0., std)
 
-class Model(nn.Module):
-	def __init__(self, *args, **kwargs):
-		super(Model, self).__init__()
-		self._record = False
-		self._merge_bn = False
-		self.is_built = False
-		self.initialize(*args, **kwargs)
-
-	def initialize(self, *args, **kwargs):
-		pass 
-
-	def build(self, *inputs):
-		pass 
-
-	def build_forward(self, *inputs, **kwargs):
-		return self.forward(*inputs, **kwargs)
-
-	def __call__(self, *input, **kwargs):
-		if not self.is_built:
-			self.build(*input)
-		for hook in self._forward_pre_hooks.values():
-			result = hook(self, input)
-			if result is not None:
-				if not isinstance(result, tuple):
-					result = (result,)
-				input = result
-		if torch._C._get_tracing_state():
-			result = self._slow_forward(*input, **kwargs)
-		else:
-			if not self.is_built:
-				result = self.build_forward(*input, **kwargs)
-			else:
-				result = self.forward(*input, **kwargs)
-		for hook in self._forward_hooks.values():
-			hook_result = hook(self, input, result)
-			if hook_result is not None:
-				result = hook_result
-		if len(self._backward_hooks) > 0:
-			var = result
-			while not isinstance(var, torch.Tensor):
-				if isinstance(var, dict):
-					var = next((v for v in var.values() if isinstance(v, torch.Tensor)))
-				else:
-					var = var[0]
-			grad_fn = var.grad_fn
-			if grad_fn is not None:
-				for hook in self._backward_hooks.values():
-					wrapper = functools.partial(hook, self)
-					functools.update_wrapper(wrapper, hook)
-					grad_fn.register_hook(wrapper)
-		self.is_built = True
-		return result
-
-	def record(self):
-		def set_record_flag(obj):
-			obj._record = True
-		self.apply(set_record_flag)
-
-	def un_record(self):
-		def unset_record_flag(obj):
-			obj._record = False
-		self.apply(unset_record_flag)
-
-	def merge_bn(self):
-		def set_merge_bn(obj):
-			obj._merge_bn = True 
-		self.apply(set_merge_bn)
-
-	def bn_eps(self, value):
-		def set_eps(obj):
-			obj.eps = value
-		self.apply(set_eps)
+######  Layers 
 
 class conv2D(Model):
 	def initialize(self, size, outchn, stride=1, pad='SAME_LEFT', dilation_rate=1, usebias=True, gropus=1):
@@ -100,10 +32,8 @@ class conv2D(Model):
 	def _parse_args(self, input_shape):
 		inchannel = input_shape[1]
 		self.inchannel = inchannel
-		# print('INC', inchannel)
 		# parse args
 		if isinstance(self.size,list):
-			# self.size = [self.size[0],self.size[1],inchannel,self.outchn]
 			if self.pad == 'VALID':
 				self.pad = 0
 			else:
@@ -117,9 +47,7 @@ class conv2D(Model):
 			self.size = [self.outchn, inchannel // self.gropus, self.size, self.size]
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
-		# self.inchannel = inp
 		self._parse_args(inp.shape)
 		self.weight = Parameter(torch.Tensor(*self.size))
 		if self.usebias:
@@ -128,18 +56,44 @@ class conv2D(Model):
 			self.register_parameter('bias', None)
 		self.reset_params()
 
+		if self._quant:
+			bit_type = self.get_flag('QActBit')
+			if bit_type is None:
+				bit_type = 'int8'
+			obs_type = self.get_flag('QActObserver')
+			if obs_type is None:
+				obs_type = 'minmax'
+			self.input_quantizer = QQuantizers['uniform'](zero_offset=False, bit_type=bit_type, observer=obs_type)
+			self.w_quantizer = QQuantizers['uniform'](zero_offset=True, mode='channel_wise', is_weight=True)
+
 	def reset_params(self):
-		_resnet_normal(self.weight)
+		if self.get_flag('conv_init_mode')=='normal':
+			init.normal_(self.weight, std=0.001)
+		elif self.get_flag('conv_init_mode')=='resnet':
+			_resnet_normal(self.weight)
+		else:
+			init.kaiming_uniform_(self.weight, a=math.sqrt(5))
 		if self.bias is not None:
-			fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
-			bound = 1 / math.sqrt(fan_in)
-			init.uniform_(self.bias, -bound, bound)
+			if self.get_flag('conv_init_mode')=='normal':
+				init.zeros_(self.bias)
+			elif self.get_flag('conv_init_mode')=='resnet':
+				fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
+				bound = 1 / math.sqrt(fan_in)
+				init.uniform_(self.bias, -bound, bound)
+			else:
+				fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
+				if fan_in != 0:
+					bound = 1 / math.sqrt(fan_in)
+					init.uniform_(self.bias, -bound, bound)
 
 	def forward(self, x):
-		return F.conv2d(x, self.weight, self.bias, self.stride, self.pad, self.dilation_rate, self.gropus)
+		weight = self.weight
+		if self._quant:
+			x = self.input_quantizer(x)
+			weight = self.w_quantizer(weight)
+		return F.conv2d(x, weight, self.bias, self.stride, self.pad, self.dilation_rate, self.gropus)
 
 	def to_torch(self):
-		# print('inchannel',self.inchannel)
 		conv = nn.Conv2d(in_channels = self.inchannel, out_channels = self.outchn, kernel_size = tuple(self.size[2:]), stride = self.stride,\
 						padding = self.pad, padding_mode = 'zeros', dilation = self.dilation_rate, groups = self.gropus, bias = self.usebias)
 		conv.weight.data[:] = self.weight.data[:]
@@ -147,6 +101,7 @@ class conv2D(Model):
 			conv.bias.data[:] = self.bias.data[:]
 		return conv 
 
+
 class deconv2D(Model):
 	def initialize(self, size, outchn, stride=1, pad='SAME_LEFT', dilation_rate=1, usebias=True, gropus=1):
 		self.size = size
@@ -165,16 +120,15 @@ class deconv2D(Model):
 		if isinstance(self.size,int):
 			if self.pad == 'VALID':
 				self.pad = 0
+				self.out_pad = 0
 			else:
 				self.pad = (self.size + (self.dilation_rate-1) * ( self.size-1 ))//2 - (1 - self.size%2)
-				# self.pad = self.dilation_rate * (self.size - 1 ) 
-			self.out_pad = self.stride - 1
+				self.out_pad = self.stride - 1
 			self.size = [inchannel, self.outchn // self.gropus, self.size, self.size]
 		else:
 			raise Exception("Deconv kernel only supports int")
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 		self.weight = Parameter(torch.Tensor(*self.size))
@@ -184,6 +138,16 @@ class deconv2D(Model):
 			self.register_parameter('bias', None)
 		self.reset_params()
 
+		if self._quant:
+			bit_type = self.get_flag('QActBit')
+			if bit_type is None:
+				bit_type = 'int8'
+			obs_type = self.get_flag('QActObserver')
+			if obs_type is None:
+				obs_type = 'minmax'
+			self.input_quantizer = QQuantizers['uniform'](zero_offset=False, bit_type=bit_type, observer=obs_type)
+			self.w_quantizer = QQuantizers['uniform'](zero_offset=True, mode='channel_wise', is_weight=True)
+
 	def reset_params(self):
 		_resnet_normal(self.weight)
 		if self.bias is not None:
@@ -193,13 +157,18 @@ class deconv2D(Model):
 
 	def forward(self, x):
 		inh, inw = x.shape[2], x.shape[3]
-		x = F.conv_transpose2d(x, self.weight, self.bias, self.stride, self.pad, self.out_pad, self.gropus, self.dilation_rate)
+		weight = self.weight
+		if self._quant:
+			x = self.input_quantizer(x)
+			weight = self.w_quantizer(weight)
+		x = F.conv_transpose2d(x, weight, self.bias, self.stride, self.pad, self.out_pad, self.gropus, self.dilation_rate)
 		outh, outw = x.shape[2], x.shape[3]
 		if self.padmethod=='SAME_LEFT':
 			if outh!=inh*self.stride or outw!=inw*self.stride:
 				x = x[:,:,:inh*self.stride,:inw*self.stride]
 		return x 
 
+
 class dwconv2D(Model):
 	# depth-wise conv2d
 	def initialize(self, size, multiplier, stride=1, pad='SAME_LEFT', dilation_rate=1, usebias=True):
@@ -218,7 +187,6 @@ class dwconv2D(Model):
 		self.outchn = self.multiplier * inchannel
 		# parse args
 		if isinstance(self.size,list):
-			# self.size = [self.size[0],self.size[1],inchannel,self.outchn]
 			if self.pad == 'VALID':
 				self.pad = 0
 			else:
@@ -232,7 +200,6 @@ class dwconv2D(Model):
 			self.size = [self.multiplier * inchannel, 1, self.size, self.size]
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 		self.weight = Parameter(torch.Tensor(*self.size))
@@ -260,6 +227,7 @@ class dwconv2D(Model):
 			conv.bias.data[:] = self.bias.data[:]
 		return conv 
 
+
 class conv1D(Model):
 	def initialize(self, size, outchn, stride=1, pad='SAME_LEFT', dilation_rate=1, usebias=True, gropus=1):
 		self.size = size
@@ -281,7 +249,6 @@ class conv1D(Model):
 		self.size = [self.outchn, inchannel // self.gropus, self.size]
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 		self.weight = Parameter(torch.Tensor(*self.size))
@@ -301,6 +268,7 @@ class conv1D(Model):
 	def forward(self, x):
 		return F.conv1d(x, self.weight, self.bias, self.stride, self.pad, self.dilation_rate, self.gropus)
 
+
 class conv3D(Model):
 	def initialize(self, size, outchn, stride=1, pad='SAME_LEFT', dilation_rate=1, usebias=True, gropus=1):
 		self.size = size
@@ -316,7 +284,6 @@ class conv3D(Model):
 		inchannel = input_shape[1]
 		# parse args
 		if isinstance(self.size,list) or isinstance(self.size, tuple):
-			# self.size = [self.size[0],self.size[1],inchannel,self.outchn]
 			if self.pad == 'VALID':
 				self.pad = 0
 			else:
@@ -330,7 +297,6 @@ class conv3D(Model):
 			self.size = [self.outchn, inchannel // self.gropus, self.size, self.size, self.size]
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 		self.weight = Parameter(torch.Tensor(*self.size))
@@ -350,6 +316,7 @@ class conv3D(Model):
 	def forward(self, x):
 		return F.conv3d(x, self.weight, self.bias, self.stride, self.pad, self.dilation_rate, self.gropus)
 
+
 class fclayer(Model):
 	def initialize(self, outsize, usebias=True, norm=False):
 		self.outsize = outsize
@@ -357,8 +324,7 @@ class fclayer(Model):
 		self.norm = norm
 
 	def build(self, *inputs):
-		# print('building...')
-		self.insize = inputs[0].shape[1]
+		self.insize = inputs[0].shape[-1]
 		self.weight = Parameter(torch.Tensor(self.outsize, self.insize))
 		if self.usebias:
 			self.bias = Parameter(torch.Tensor(self.outsize))
@@ -366,16 +332,28 @@ class fclayer(Model):
 			self.register_parameter('bias', None)
 		self.reset_params()
 
+		if self._quant:
+			bit_type = self.get_flag('QActBit')
+			if bit_type is None:
+				bit_type = 'int8'
+			obs_type = self.get_flag('QActObserver')
+			if obs_type is None:
+				obs_type = 'minmax'
+			self.input_quantizer = QQuantizers['uniform'](zero_offset=False, bit_type=bit_type, observer=obs_type)
+			self.w_quantizer = QQuantizers['uniform'](zero_offset=True, mode='channel_wise', is_weight=True)
+
+
 	def reset_params(self):
-		# init.kaiming_uniform_(self.weight, a=math.sqrt(5))
-		init.normal_(self.weight, std=0.001)
-		# _resnet_normal(self.weight)
-		# print('Reset fc params...')
-		if self.bias is not None:
-		# 	fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
-		# 	bound = 1 / math.sqrt(fan_in)
-		# 	init.uniform_(self.bias, -bound, bound)
-			init.zeros_(self.bias)
+		if self.get_flag('fc_init_mode')=='normal':
+			init.normal_(self.weight, std=0.001)
+			if self.bias is not None:
+				init.zeros_(self.bias)
+		else:
+			init.kaiming_uniform_(self.weight, a=math.sqrt(5))
+			if self.bias is not None:
+				fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
+				bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
+				init.uniform_(self.bias, -bound, bound)
 
 	def forward(self, x):
 		if self.norm:
@@ -386,6 +364,10 @@ class fclayer(Model):
 			weight = self.weight / wnorm
 		else:
 			weight = self.weight
+
+		if self._quant:
+			x = self.input_quantizer(x).contiguous()
+			weight = self.w_quantizer(weight).contiguous()
 		return F.linear(x, weight, self.bias)
 
 	def to_torch(self):
@@ -395,13 +377,6 @@ class fclayer(Model):
 			fc.bias.data[:] = self.bias.data[:]
 		return fc 
 
-def flatten(x):
-	x = x.reshape(x.size(0), -1)
-	return x 
-
-class Flatten(Model):
-	def forward(self, x):
-		return flatten(x)
 
 class MaxPool2d(Model):
 	def initialize(self, size, stride=1, pad='SAME_LEFT', dilation_rate=1):
@@ -414,7 +389,6 @@ class MaxPool2d(Model):
 		inchannel = input_shape[1]
 		# parse args
 		if isinstance(self.size,list) or isinstance(self.size, tuple):
-			# self.size = [self.size[0],self.size[1],inchannel,self.outchn]
 			if self.pad == 'VALID':
 				self.pad = 0
 			else:
@@ -426,13 +400,13 @@ class MaxPool2d(Model):
 				self.pad = self.size//2
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 
 	def forward(self, x):
 		return F.max_pool2d(x, self.size, self.stride, self.pad, self.dilation_rate, False, False)
 
+
 class AvgPool2d(Model):
 	def initialize(self, size, stride=1, pad='SAME_LEFT'):
 		self.size = size
@@ -443,7 +417,6 @@ class AvgPool2d(Model):
 		inchannel = input_shape[1]
 		# parse args
 		if isinstance(self.size,list) or isinstance(self.size, tuple):
-			# self.size = [self.size[0],self.size[1],inchannel,self.outchn]
 			if self.pad == 'VALID':
 				self.pad = 0
 			else:
@@ -455,20 +428,20 @@ class AvgPool2d(Model):
 				self.pad = self.size//2
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 
 	def forward(self, x):
 		return F.avg_pool2d(x, self.size, self.stride, self.pad, False, True)
 
+
 class BatchNorm(Model):
 	# _version = 2
 	# __constants__ = ['track_running_stats', 'momentum', 'eps', 'weight', 'bias',
 	# 				 'running_mean', 'running_var', 'num_batches_tracked',
 	# 				 'num_features', 'affine', 'weight', 'bias']
 
-	def initialize(self, eps=2e-5, momentum=0.01, affine=True,
+	def initialize(self, eps=None, momentum=0.01, affine=True,
 				 track_running_stats=True):
 		self.eps = eps
 		self.momentum = momentum
@@ -476,7 +449,6 @@ class BatchNorm(Model):
 		self.track_running_stats = track_running_stats
 		
 	def build(self, *inputs):
-		# print('building...')
 		num_features = inputs[0].shape[1]
 		self.num_features = num_features
 		if self.affine:
@@ -494,6 +466,8 @@ class BatchNorm(Model):
 			self.register_parameter('running_var', None)
 			self.register_parameter('num_batches_tracked', None)
 		self.reset_parameters()
+		eps = self.get_flag('bn_eps')
+		self.eps = 2e-5 if eps is None else eps 
 
 	def reset_running_stats(self):
 		if self.track_running_stats:
@@ -508,7 +482,6 @@ class BatchNorm(Model):
 			init.zeros_(self.bias)
 
 	def forward(self, input):
-		global record_params
 		if self.momentum is None:
 			exponential_average_factor = 0.0
 		else:
@@ -527,15 +500,6 @@ class BatchNorm(Model):
 			input, self.running_mean, self.running_var, self.weight, self.bias,
 			self.training or not self.track_running_stats,
 			exponential_average_factor, self.eps)
-		if hasattr(self, '_record'):
-			if self._record:
-				res = {}
-				for p in self.named_parameters():
-					res[p[0]] = p[1]
-				for p in self.named_buffers():
-					res[p[0]] = p[1]
-				record_params.append(res)
-			self.un_record()
 		return result
 
 	def to_torch(self):
@@ -563,13 +527,35 @@ class BatchNorm(Model):
 	# 		state_dict, prefix, local_metadata, strict,
 	# 		missing_keys, unexpected_keys, error_msgs)
 
-def GlobalAvgPool2D(x):
-	x = x.mean(dim=(2,3), keepdim=True)
-	return x 
 
-class GlobalAvgPool2DLayer(Model):
+class LayerNorm(Model):
+	def initialize(self, n_dims_to_keep, affine=True, eps=1e-5):
+		self.n_dims_to_keep = n_dims_to_keep
+		self.affine = affine
+		self.eps = eps 
+
+	def build(self, *inputs):
+		shape = inputs[0].shape
+		n_dims = len(shape)
+		self.dim_to_reduce = tuple(range(n_dims - self.n_dims_to_keep, n_dims))
+		if self.affine:
+			self.weight = Parameter(torch.Tensor(*shape[-self.n_dims_to_keep:]))
+			self.bias = Parameter(torch.Tensor(*shape[-self.n_dims_to_keep:]))
+		self.reset_params()
+
+	def reset_params(self):
+		if self.affine:
+			init.ones_(self.weight)
+			init.zeros_(self.bias)
+
 	def forward(self, x):
-		return GlobalAvgPool2D(x)
+		mean = torch.mean(x, dim=self.dim_to_reduce, keepdim=True)
+		std = torch.std(x, dim=self.dim_to_reduce, keepdim=True)
+		if self.affine:
+			return self.weight * (x - mean) / (std + self.eps) + self.bias
+		else:
+			return (x - mean) / (std + self.eps)
+
 
 class NNUpSample(Model):
 	def initialize(self, scale):
@@ -580,7 +566,6 @@ class NNUpSample(Model):
 		self.size = [self.inchannel, 1, self.scale, self.scale]
 
 	def build(self, *inputs):
-		# print('building...')
 		inp = inputs[0]
 		self._parse_args(inp.shape)
 		self.weight = Parameter(torch.Tensor(*self.size), requires_grad=False)
@@ -595,33 +580,39 @@ class NNUpSample(Model):
 		w = self.weight
 		return F.conv_transpose2d(x, w, self.bias, self.scale, 0, 0, self.inchannel, 1)
 
+
 def activation(x, act, **kwargs):
 	if act==-1:
 		return x
 	elif act==0:
 		return F.relu(x)
 	elif act==1:
-		return F.leaky_relu(x, negative_slope=0.2)
+		return F.leaky_relu(x, negative_slope=0.1)
 	elif act==2:
 		return F.elu(x)
 	elif act==3:
 		return F.tanh(x)
 	elif act==6:
 		return torch.sigmoid(x)
+	elif act==10:
+		return F.gelu(x)
+
 
 class Activation(Model):
 	def initialize(self, act):
 		self.act = act 
 		if act==8:
-			self.act = torch.nn.PReLU(num_parameters=outchn)
+			self.act = torch.nn.PReLU(num_parameters=outchn)  # this line is buggy
 		elif act==9:
 			self.act = torch.nn.PReLU(num_parameters=1)
+
 	def forward(self, x):
 		if self.act==8 or self.act==9:
 			return self.act(x)
 		else:
 			return activation(x, self.act)
 
+
 class graphConvLayer(Model):
 	def initialize(self, outsize, usebias=True, norm=True):
 		self.outsize = outsize
@@ -671,10 +662,12 @@ class graphConvLayer(Model):
 		res = F.linear(res, self.weight, self.bias)
 		return res 
 	
+
 class BilinearUpSample(Model):
 	def initialize(self, factor):
 		self.factor = factor
 		self.pad0 = factor//2 * 3 + factor%2
+
 	def build(self, *inputs):
 		inp = inputs[0]
 		self.inchn = inp.shape[1]
@@ -684,6 +677,7 @@ class BilinearUpSample(Model):
 		k = k[None, ...]
 		k = np.repeat(k, self.inchn, axis=0)
 		self.weight = Parameter(torch.from_numpy(k), requires_grad=False)
+
 	def upsample_kernel(self,size):
 		factor = (size +1)//2
 		if size%2==1:
@@ -693,11 +687,63 @@ class BilinearUpSample(Model):
 		og = np.ogrid[:size, :size]
 		k = (1 - abs(og[0]-center)/factor) * (1-abs(og[1]-center)/factor)
 		return np.float32(k)
+
 	def forward(self, x):
 		x = F.pad(x, (1,1,1,1), 'replicate')
 		x = F.conv_transpose2d(x, self.weight, None, self.factor, self.pad0, 0, self.inchn, 1)
 		return x 
+
 	def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
 		pass
+
 	def _save_to_state_dict(self, destination, prefix, keep_vars):
-		pass
\ No newline at end of file
+		pass
+
+
+class DeformConv2D(Model):
+	def initialize(self, size, outchn, stride=1, pad='SAME_LEFT', dilation_rate=1, usebias=True):
+		self.size = size
+		self.outchn = outchn
+		self.stride = stride
+		self.usebias = usebias
+		self.dilation_rate = dilation_rate
+		assert (pad in ['VALID','SAME_LEFT'])
+		self.pad = pad 
+
+	def _parse_args(self, input_shape):
+		inchannel = input_shape[1]
+		self.inchannel = inchannel
+		# parse args
+		if isinstance(self.size,list):
+			if self.pad == 'VALID':
+				self.pad = 0
+			else:
+				self.pad = ((self.size[0]+ (self.dilation_rate-1) * ( self.size-1 ))//2, (self.size[1]+ (self.dilation_rate-1) * ( self.size-1 ))//2)
+			self.size = [self.outchn, inchannel, self.size[0], self.size[1]]
+		else:
+			if self.pad == 'VALID':
+				self.pad = 0
+			else:
+				self.pad = (self.size + (self.dilation_rate-1) * ( self.size-1 ))//2
+			self.size = [self.outchn, inchannel, self.size, self.size]
+
+	def build(self, *inputs):
+		inp = inputs[0]
+		self._parse_args(inp.shape)
+		self.weight = Parameter(torch.Tensor(*self.size))
+		if self.usebias:
+			self.bias = Parameter(torch.Tensor(self.outchn))
+		else:
+			self.register_parameter('bias', None)
+		self.reset_params()
+
+	def reset_params(self):
+		_resnet_normal(self.weight)
+		if self.bias is not None:
+			fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
+			bound = 1 / math.sqrt(fan_in)
+			init.uniform_(self.bias, -bound, bound)
+
+	def forward(self, x, offset):
+		return ops.deform_conv2d(x, offset, self.weight, self.bias, self.stride, self.pad, self.dilation_rate)
+
